# -*- coding: utf-8 -*-

import pytest

from goodplay_helpers import smart_create

pytestmark = pytest.mark.integration


def test_passed_on_selfcontained_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role1/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role1/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role1

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_failed_on_selfcontained_role_without_meta(testdir):
    smart_create(testdir.tmpdir, '''
    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role1/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role1/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role1

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_role_beside_with_same_name_as_dependent_role_is_not_used(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1external.run"
        state: touch

    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1local.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1external not run
          file:
            path: "{{ playbook_dir }}/.role1external.run"
            state: file
          tags: test

        - name: assert role1local run
          file:
            path: "{{ playbook_dir }}/.role1local.run"
            state: absent
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=3)


def test_roles_beside_with_same_names_as_dependent_roles_are_not_used(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1external.run"
        state: touch

    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1local.run"
        state: touch

    ## external-role-base/role2.tar.gz
    #### role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    #### role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2external.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2local.run"
        state: touch

    ## local-role-base/role3/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role2
        src: external-role-base/role2.tar.gz

    ## local-role-base/role3/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role3.run"
        state: touch

    ## local-role-base/role3/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role3/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role3

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1external not run
          file:
            path: "{{ playbook_dir }}/.role1external.run"
            state: file
          tags: test

        - name: assert role1local run
          file:
            path: "{{ playbook_dir }}/.role1local.run"
            state: absent
          tags: test

        - name: assert role2external not run
          file:
            path: "{{ playbook_dir }}/.role2external.run"
            state: file
          tags: test

        - name: assert role2local run
          file:
            path: "{{ playbook_dir }}/.role2local.run"
            state: absent
          tags: test

        - name: assert role3 run
          file:
            path: "{{ playbook_dir }}/.role3.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=5)


def test_passed_on_role_with_external_dependent_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


def test_passed_on_role_with_multi_level_external_dependent_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## external-role-base/role2.tar.gz
    #### role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    #### role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role3/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role2
        src: external-role-base/role2.tar.gz

    ## local-role-base/role3/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role3.run"
        state: touch

    ## local-role-base/role3/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role3/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role3

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test

        - name: assert role3 run
          file:
            path: "{{ playbook_dir }}/.role3.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=3)


def test_passed_on_role_with_external_soft_dependent_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/requirements.yml
    - name: role1
      src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role1
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


def test_passed_on_role_with_multi_level_external_soft_dependent_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## external-role-base/role2.tar.gz
    #### role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    #### role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role3/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role3/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role3.run"
        state: touch

    ## local-role-base/role3/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role3/tests/requirements.yml
    - name: role2
      src: external-role-base/role2.tar.gz

    ## local-role-base/role3/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2
        - role: role3

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test

        - name: assert role3 run
          file:
            path: "{{ playbook_dir }}/.role3.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=3)


def test_soft_dependency_takes_precedence_over_role_dependency(testdir):
    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## external-role-base/role1soft.tar.gz
    #### role1soft/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1soft/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1soft.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/requirements.yml
    - name: role1
      src: external-role-base/role1soft.tar.gz

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 not run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: absent
          tags: test

        - name: assert role1soft run
          file:
            path: "{{ playbook_dir }}/.role1soft.run"
            state: file
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=3)


def test_failed_on_unresolvable_role(testdir):
    smart_create(testdir.tmpdir, '''
    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role1/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role1/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


# feature: use local roles

def test_when_use_local_roles_is_enabled_ansible_path_takes_precedence_over_soft_dependency(
        testdir, monkeypatch):
    monkeypatch.setattr(
        'ansible.constants.DEFAULT_ROLES_PATH',
        [testdir.tmpdir.join('local-role-base').strpath])

    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1soft.tar.gz
    #### role1soft/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1soft/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1soft.run"
        state: touch

    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/requirements.yml
    - name: role1
      src: external-role-base/role1soft.tar.gz

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role1
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test

        - name: assert role1soft not run
          file:
            path: "{{ playbook_dir }}/.role1soft.run"
            state: absent
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s', '--use-local-roles')
    result.assertoutcome(passed=3)


def test_when_use_local_roles_is_enabled_ansible_path_takes_precedence_over_role_dependency(
        testdir, monkeypatch):
    monkeypatch.setattr(
        'ansible.constants.DEFAULT_ROLES_PATH',
        [testdir.tmpdir.join('local-role-base').strpath])

    smart_create(testdir.tmpdir, '''
    ## external-role-base/role1.tar.gz
    #### role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    #### role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1external.run"
        state: touch

    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1local.run"
        state: touch

    ## local-role-base/role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - name: role1
        src: external-role-base/role1.tar.gz

    ## local-role-base/role2/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role2.run"
        state: touch

    ## local-role-base/role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## local-role-base/role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1local run
          file:
            path: "{{ playbook_dir }}/.role1local.run"
            state: file
          tags: test

        - name: assert role1external not run
          file:
            path: "{{ playbook_dir }}/.role1external.run"
            state: absent
          tags: test

        - name: assert role2 run
          file:
            path: "{{ playbook_dir }}/.role2.run"
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s', '--use-local-roles')
    result.assertoutcome(passed=3)


def test_when_use_local_roles_is_enabled_ansible_path_is_considered(testdir, monkeypatch):
    monkeypatch.setenv(
        'ANSIBLE_ROLES_PATH',
        testdir.tmpdir.join('some', 'ansible', 'roles').strpath)

    smart_create(testdir.tmpdir, '''
    ## some/ansible/roles/role1/tasks/main.yml
    - file:
        path: "{{ playbook_dir }}/.role1.run"
        state: touch

    ## role2/meta/main.yml
    galaxy_info:
      author: John Doe
    dependencies:
      - role1

    ## role2/tasks/main.yml
    - ping:

    ## role2/tests/inventory
    127.0.0.1 ansible_connection=local

    ## role2/tests/test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      roles:
        - role: role2

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert role1 run
          file:
            path: "{{ playbook_dir }}/.role1.run"
            state: file
          tags: test
    ''')

    # explicitly run with environment variable and subprocess to ensure
    # correct integration with Ansible
    result = testdir.runpytest_subprocess('-s', '--use-local-roles')
    result.assert_outcomes(passed=1)



# -*- coding: utf-8 -*-

# This connection plugin has been implemented as the docker connection plugin
# provided by Ansible does not support become_user with Docker's native user
# management (as of version 2.1.0). The necessary changes have been
# implemented in PR ansible/ansible#15556 but not accepted.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import ansible.constants as C
from ansible.playbook.play_context import PlayContext
import ansible.plugins.connection.docker


def monkeypatch_play_context():
    original_make_become_cmd = PlayContext.make_become_cmd

    def make_become_cmd_with_dockerexec(self, cmd, executable=None):
        if self.become and self.become_method == 'dockerexec':
            self.prompt = None
            return cmd

        return original_make_become_cmd(self, cmd, executable)

    PlayContext.make_become_cmd = make_become_cmd_with_dockerexec


monkeypatch_play_context()


class Connection(ansible.plugins.connection.docker.Connection):
    transport = 'goodplaydocker'

    become_methods = set(C.BECOME_METHODS).difference({'su'}).union({'dockerexec'})

    def _build_exec_cmd(self, cmd):
        local_cmd = [self.docker_cmd]

        if self._play_context.docker_extra_args:
            local_cmd += self._play_context.docker_extra_args.split(' ')

        local_cmd += ['exec']

        if self._play_context.become and self._play_context.become_method == 'dockerexec':
            exec_user = self._play_context.become_user
        else:
            exec_user = self.remote_user

        if exec_user is not None:
            local_cmd += ['-u', exec_user]

        # -i is needed to keep stdin open which allows pipelining to work
        local_cmd += ['-i', self._play_context.remote_addr] + cmd

        return local_cmd



# -*- coding: utf-8 -*-

from goodplay_helpers import skip_if_no_docker, smart_create


@skip_if_no_docker
def test_become_user_on_task_without_become_does_not_execute_as_become_user(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      host1:
        image: centos:centos6
        tty: True

    ## inventory
    host1 ansible_user=root

    ## test_playbook.yml
    - hosts: host1
      gather_facts: no
      tasks:
        - name: create system group myservice
          group:
            name: myservice
            system: yes
            state: present

        - name: create system user myservice
          user:
            name: myservice
            group: myservice
            shell: /sbin/nologin
            system: yes
            state: present

        - name: create myservice directory
          file:
            path: /opt/myservice
            owner: myservice
            group: myservice
            mode: 0700
            state: directory

        - name: intentionally only specify become_user on this task
          file:
            path: /opt/myservice/somefile
            state: touch
          become_user: myservice

        - name: ensure somefile is owned by root user which is not the become_user
          file:
            path: /opt/myservice/somefile
            owner: root
            group: root
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_become_with_become_user_on_play(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      host1:
        image: centos:centos6
        tty: True

    ## inventory
    host1 ansible_user=root

    ## test_playbook.yml
    - hosts: host1
      gather_facts: no
      become_user: myservice
      tasks:
        - name: create system group myservice
          group:
            name: myservice
            system: yes
            state: present

        - name: create system user myservice
          user:
            name: myservice
            group: myservice
            shell: /sbin/nologin
            system: yes
            state: present

        - name: create myservice directory
          file:
            path: /opt/myservice
            owner: myservice
            group: myservice
            mode: 0700
            state: directory

        - name: make some file operation as myservice user
          file:
            path: /opt/myservice/somefile
            state: touch
          become: yes

        - name: ensure somefile is owned by myservice user
          file:
            path: /opt/myservice/somefile
            owner: myservice
            group: myservice
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_become_with_become_user_on_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      host1:
        image: centos:centos6
        tty: True

    ## inventory
    host1 ansible_user=root

    ## test_playbook.yml
    - hosts: host1
      gather_facts: no
      tasks:
        - name: create system group myservice
          group:
            name: myservice
            system: yes
            state: present

        - name: create system user myservice
          user:
            name: myservice
            group: myservice
            shell: /sbin/nologin
            system: yes
            state: present

        - name: create myservice directory
          file:
            path: /opt/myservice
            owner: myservice
            group: myservice
            mode: 0700
            state: directory

        - name: make some file operation as myservice user
          file:
            path: /opt/myservice/somefile
            state: touch
          become: yes
          become_user: myservice

        - name: ensure somefile is owned by myservice user
          file:
            path: /opt/myservice/somefile
            owner: myservice
            group: myservice
            state: file
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)



# -*- coding: utf-8 -*-

import tarfile

from goodplay_helpers import smart_create


def test_smart_create_empty(tmpdir):
    smart_create(tmpdir, '')

    assert tmpdir.listdir() == []


def test_smart_create_linebreaks_before_first_file(tmpdir):
    smart_create(tmpdir, '''

    ## file1
    Hello World
    ''')

    assert tmpdir.join('file1').check(file=True)
    assert tmpdir.join('file1').read() == 'Hello World'


def test_smart_create_single_file_empty(tmpdir):
    smart_create(tmpdir, '''
    ## file1

    ''')

    assert tmpdir.join('file1').check(file=True)
    assert tmpdir.join('file1').read() == ''


def test_smart_create_single_file_single_line(tmpdir):
    smart_create(tmpdir, '''
    ## file1
    Hello World
    ''')

    assert tmpdir.join('file1').check(file=True)
    assert tmpdir.join('file1').read() == 'Hello World'


def test_smart_create_single_file_multiple_lines(tmpdir):
    smart_create(tmpdir, '''
    ## file1
    Hello
    World
    ''')

    assert tmpdir.join('file1').check(file=True)
    assert tmpdir.join('file1').read() == 'Hello\nWorld'


def test_smart_create_single_file_in_subdir(tmpdir):
    smart_create(tmpdir, '''
    ## dir1/file1
    Hello
    World
    ''')

    assert tmpdir.join('dir1').check(dir=True)
    assert tmpdir.join('dir1', 'file1').check(file=True)
    assert tmpdir.join('dir1', 'file1').read() == 'Hello\nWorld'


def test_smart_create_multiple_files(tmpdir):
    smart_create(tmpdir, '''
    ## file1
    Hello
    World
    ## file2
    Hello, John Doe
    ''')

    assert tmpdir.join('file1').check(file=True)
    assert tmpdir.join('file1').read() == 'Hello\nWorld'

    assert tmpdir.join('file2').check(file=True)
    assert tmpdir.join('file2').read() == 'Hello, John Doe'


def test_smart_create_complex(tmpdir):
    smart_create(tmpdir, '''
    ## VERSION
    1.0.1

    ## docs/README.md
    # My project

    ## helloworld/main.py
    print 'Hello World'

    ''')

    assert tmpdir.join('VERSION').read() == '1.0.1\n'
    assert tmpdir.join('docs', 'README.md').read() == '# My project\n'
    assert tmpdir.join('helloworld', 'main.py').read() == "print 'Hello World'\n"


def test_smart_create_single_archive_single_file(tmpdir):
    smart_create(tmpdir, '''
    ## archive.tar.gz
    #### file1
    Hello World
    ''')

    archive_path = tmpdir.join('archive.tar.gz')
    extracted_path = tmpdir.join('archive.tar.gz.extracted')
    extracted_path.ensure(dir=True)

    with tarfile.open(str(archive_path), 'r:gz') as tar:
        tar.extractall(str(extracted_path))

    assert extracted_path.join('file1').read() == 'Hello World'


def test_smart_create_single_archive_multiple_files(tmpdir):
    smart_create(tmpdir, '''
    ## archive.tar.gz
    #### file1
    Hello World
    #### file2
    Hello
    World
    ''')

    archive_path = tmpdir.join('archive.tar.gz')
    extracted_path = tmpdir.join('archive.tar.gz.extracted')
    extracted_path.ensure(dir=True)

    with tarfile.open(str(archive_path), 'r:gz') as tar:
        tar.extractall(str(extracted_path))

    assert extracted_path.join('file1').read() == 'Hello World'
    assert extracted_path.join('file2').read() == 'Hello\nWorld'


def test_smart_create_single_archive_in_subdir(tmpdir):
    smart_create(tmpdir, '''
    ## dir1/archive.tar.gz
    #### file1
    Hello World
    ''')

    assert tmpdir.join('dir1').check(dir=True)

    archive_path = tmpdir.join('dir1', 'archive.tar.gz')
    extracted_path = tmpdir.join('archive.tar.gz.extracted')
    extracted_path.ensure(dir=True)

    with tarfile.open(str(archive_path), 'r:gz') as tar:
        tar.extractall(str(extracted_path))

    assert extracted_path.join('file1').read() == 'Hello World'


def test_smart_create_single_archive_with_subdirs(tmpdir):
    smart_create(tmpdir, '''
    ## archive.tar.gz
    #### dir1/dir2/hello_world.py
    print 'Hello World'
    #### another/README
    Some text
    ''')

    archive_path = tmpdir.join('archive.tar.gz')
    extracted_path = tmpdir.join('archive.tar.gz.extracted')
    extracted_path.ensure(dir=True)

    with tarfile.open(str(archive_path), 'r:gz') as tar:
        tar.extractall(str(extracted_path))

    assert extracted_path.join('dir1', 'dir2', 'hello_world.py').read() == "print 'Hello World'"
    assert extracted_path.join('another', 'README').read() == 'Some text'


def test_smart_create_multiple_archives_multiple_files(tmpdir):
    smart_create(tmpdir, '''
    ## archive1.tar.gz
    #### file1
    Hello World
    ## archive2.tar.gz
    #### file2
    Content of file2
    ''')

    # extract archive1.tar.gz
    archive1_path = tmpdir.join('archive1.tar.gz')
    extracted1_path = tmpdir.join('archive1.tar.gz.extracted')
    extracted1_path.ensure(dir=True)

    with tarfile.open(str(archive1_path), 'r:gz') as tar:
        tar.extractall(str(extracted1_path))

    assert extracted1_path.join('file1').read() == 'Hello World'

    # extract archive2.tar.gz
    archive2_path = tmpdir.join('archive2.tar.gz')
    extracted2_path = tmpdir.join('archive2.tar.gz.extracted')
    extracted2_path.ensure(dir=True)

    with tarfile.open(str(archive2_path), 'r:gz') as tar:
        tar.extractall(str(extracted2_path))

    assert extracted2_path.join('file2').read() == 'Content of file2'



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from cached_property import cached_property
from compose.cli.command import get_project
from compose.service import BuildError
from pytest import fail


def is_docker_compose_file(path):
    is_file = path.check(file=True)
    uses_docker_compose_naming = (
        path.basename.startswith('docker-compose.') and
        path.basename.endswith('.yml') and
        '..' not in path.basename)

    return is_file and uses_docker_compose_naming


def config_name_self_and_above(config_name, suffix=None):
    splitted_config_name = config_name.split('.')
    prefix = [splitted_config_name[0]]
    if not suffix:
        suffix = [splitted_config_name[-1]]
    environment_parts = splitted_config_name[1:-1]

    if environment_parts and environment_parts[-1] == 'override':
        environment_parts.pop()

    while environment_parts:
        yield '.'.join(prefix + environment_parts + suffix)
        environment_parts.pop()

    yield '.'.join(prefix + environment_parts + suffix)


def environment_names_for_playbook_path(playbook_path):
    return [environment_name_for_config_path(config_path)
            for config_path in config_paths_for_playbook_path(playbook_path)]


def config_path_for_environment_name(playbook_path, environment_name):
    for config_path in config_paths_for_playbook_path(playbook_path):
        if environment_name_for_config_path(config_path) == environment_name:
            return config_path


def config_paths_for_playbook_path(playbook_path):  # noqa: R701
    base_path = playbook_path.dirpath()

    # get relative config names sorted descending by len
    config_names = [path.relto(base_path) for path in base_path.listdir(is_docker_compose_file)]
    config_names.sort()
    config_names.sort(key=len, reverse=True)

    config_names_set = set(config_names)

    config_paths = []
    already_visited = set()

    for config_name in config_names:
        if config_name in already_visited:
            continue

        config_path = []
        continue_adding_config_paths = True

        for potential_config_override_name, potential_config_name in zip(
                config_name_self_and_above(config_name, suffix=['override', 'yml']),
                config_name_self_and_above(config_name)):
            if continue_adding_config_paths and potential_config_override_name in config_names_set:
                config_path.insert(0, potential_config_override_name)
            if continue_adding_config_paths and potential_config_name in config_names_set:
                continue_adding_config_paths = False
                config_path.insert(0, potential_config_name)

            already_visited.add(potential_config_override_name)
            already_visited.add(potential_config_name)

        config_paths.append(config_path)

    return config_paths


def environment_name_for_config_path(config_path):
    if config_path:
        environment_parts = config_path[-1].split('.')[1:-1]

        if environment_parts and environment_parts[-1] == 'override':
            environment_parts.pop()

        return '.'.join(environment_parts)


class DockerRunner(object):
    def __init__(self, ctx, environment_name=None):
        self.ctx = ctx
        self.environment_name = environment_name

    @cached_property
    def project(self):
        config_path = config_path_for_environment_name(
            self.ctx.playbook_path, self.environment_name)

        return get_project(
            project_dir=self.ctx.playbook_dir_path.strpath,
            config_path=config_path,
            project_name=self.ctx.compose_project_name(self.environment_name))

    def setup(self):
        if self.environment_name is None:
            return

        try:
            # ensure image builds are up-to-date
            # do not pull as image might only be available locally
            self.project.build(pull=False)
        except BuildError as err:
            fail("building service '{0}' failed with reason '{1}'".format(
                err.service.name, err.reason))

        self.project.up()

        inventory_lines = []
        for service in self.project.services:
            container_name = service.get_container().name
            inventory_line = ' '.join((
                service.name,
                'ansible_connection="goodplaydocker"',
                'ansible_become_method="dockerexec"',
                'ansible_host="{}"'.format(container_name),
            ))
            inventory_lines.append(inventory_line)
        inventory_content = '\n'.join(inventory_lines)

        # choose name 'inventory_goodplay' as it comes after 'inventory' when compared,
        # alphanumerical, otherwise Ansible does not merge in the ansible_connection
        # information correctly and thus tries to use an SSH connection
        self.ctx.extended_inventory_path.join('inventory_goodplay').write(inventory_content)

    def teardown(self):
        if self.environment_name is None:
            return

        self.project.kill()
        self.project.down(remove_image_type=False, include_volumes=True)



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import logging
import json

import py.path

from ..utils.subprocess import run

log = logging.getLogger(__name__)


class PlaybookRunner(object):
    def __init__(self, ctx):
        self.ctx = ctx

        self.process = None
        self.skip_wait = False
        self.failures = []
        self.all_test_tasks_skipped = True

    def run_async(self):
        this_path = py.path.local(__file__)
        callback_plugin_path = this_path.dirpath('callback_plugin')
        connection_plugin_path = this_path.dirpath('connection_plugin')

        env = self.ctx.playbook.env()
        additional_env = dict(
            PYTHONUNBUFFERED='1',
            ANSIBLE_CALLBACK_PLUGINS=callback_plugin_path.strpath,
            ANSIBLE_CALLBACK_WHITELIST='goodplay',
            ANSIBLE_CONNECTION_PLUGINS=connection_plugin_path.strpath,
        )
        env.update(additional_env)

        self.process = run(
            'ansible-playbook -vvv -i {0} {1}',
            self.ctx.extended_inventory_path, self.ctx.playbook_path,
            env=env, async_=True)

        # wait for subprocess to be responsive
        self.process.wait_events()

    def wait(self):
        self.wait_for_event()

        for line in self.process.stdout:
            log.info(line[:-1])

        self.process.wait()

        if self.all_test_tasks_skipped:
            self.failures.append('all test tasks have been skipped')

    def wait_for_event(self, event_name=None, **kwargs):
        for event in self.receive_events():
            if event['event_name'] == event_name \
                    and set(kwargs.items()).issubset(set(event['data'].items())):
                return event
            elif event['event_name'] == 'error':
                error_message = event['data']['message']
                self.failures.append(error_message)
                self.skip_wait = True
                return
            else:  # pragma: no cover
                raise Exception('found unexpected goodplay event: {0!r}'.format(event))

    def receive_events(self):
        event_line_prefix = 'GOODPLAY => '
        for line in self.process.stdout:
            log.info(line[:-1])

            if line.startswith(event_line_prefix):
                event = json.loads(line[len(event_line_prefix):])
                yield event

    def wait_for_test_task(self, task):
        if self.skip_wait:
            return

        self.wait_for_event('test-task-start', name=task.name)

    def wait_for_test_task_outcome(self, task):
        if self.skip_wait:
            return

        event = self.wait_for_event('test-task-end', name=task.name)
        outcome = event['data']['outcome'] if event else 'skipped'

        if outcome != 'skipped':
            self.all_test_tasks_skipped = False

        return outcome



# -*- coding: utf-8 -*-

import pytest

from goodplay_helpers import smart_create

pytestmark = pytest.mark.integration


def test_junitxml_removes_classname_yml_extension(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: host is reachable
          ping:
          tags: test
    ''')

    junit_xml_path = testdir.tmpdir.join('junit.xml')

    testdir.inline_run('--junitxml', str(junit_xml_path))
    junit_xml_content = junit_xml_path.read()
    assert 'classname="test_playbook"' in junit_xml_content


def test_junitxml_does_not_remove_task_name_with_yml_extension(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: update appconfig.yml
          ping:
          tags: test
    ''')

    junit_xml_path = testdir.tmpdir.join('junit.xml')

    testdir.inline_run('--junitxml', str(junit_xml_path))
    junit_xml_content = junit_xml_path.read()
    assert 'name="update appconfig.yml"' in junit_xml_content



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import yaml

from .inventory import Inventory  # noqa: F401
from .playbook import Playbook  # noqa: F401


def is_test_playbook_file(path):
    has_test_prefix = path.basename.startswith('test_')

    return has_test_prefix and is_playbook_file(path)


def is_playbook_file(path):
    is_yaml_file = path.ext == '.yml'

    if is_yaml_file:
        yaml_file_content = yaml.safe_load(path.read())
        yaml_file_is_ansible_playbook = is_playbook_content(yaml_file_content)

        return yaml_file_is_ansible_playbook

    return False


def is_playbook_content(content):
    known_playbook_attributes = ('hosts', 'include', 'import_playbook')

    return isinstance(content, list) \
        and len(content) \
        and isinstance(content[0], dict) \
        and any(content[0].get(x) for x in known_playbook_attributes)



# -*- coding: utf-8 -*-

import pytest

from goodplay_helpers import skip_if_no_docker, smart_create

pytestmark = pytest.mark.integration


@skip_if_no_docker
def test_group_vars_directory_beside_inventory_file_is_incorporated(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthostname:
        image: "busybox:latest"
        tty: True

    ## inventory
    guesthostname ansible_user=root

    ## group_vars/all
    hello: world

    ## test_playbook.yml
    - hosts: guesthostname
      gather_facts: no
      tasks:
        - name: task 1
          raw: 'echo -n {{ hello }}'
          register: echo_result
          failed_when: echo_result.stdout != 'world'
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_group_vars_directory_beside_inventory_directory_is_incorporated(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthostname:
        image: "busybox:latest"
        tty: True

    ## inventory/static
    guesthostname ansible_user=root

    ## inventory/group_vars/all
    hello: world

    ## test_playbook.yml
    - hosts: guesthostname
      gather_facts: no
      tasks:
        - name: task 1
          raw: 'echo -n {{ hello }}'
          register: echo_result
          failed_when: echo_result.stdout != 'world'
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_host_vars_directory_beside_inventory_file_is_incorporated(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthostname:
        image: "busybox:latest"
        tty: True

    ## inventory
    guesthostname ansible_user=root

    ## host_vars/guesthostname
    hello: world

    ## test_playbook.yml
    - hosts: guesthostname
      gather_facts: no
      tasks:
        - name: task 1
          raw: 'echo -n {{ hello }}'
          register: echo_result
          failed_when: echo_result.stdout != 'world'
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_host_vars_directory_beside_inventory_directory_is_incorporated(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthostname:
        image: "busybox:latest"
        tty: True

    ## inventory/static
    guesthostname ansible_user=root

    ## inventory/host_vars/guesthostname
    hello: world

    ## test_playbook.yml
    - hosts: guesthostname
      gather_facts: no
      tasks:
        - name: task 1
          raw: 'echo -n {{ hello }}'
          register: echo_result
          failed_when: echo_result.stdout != 'world'
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_role_with_multiple_environments(testdir):
    smart_create(testdir.tmpdir, '''
    ## local-role-base/role1/meta/main.yml
    galaxy_info:
      author: John Doe
      platforms:
        - name: EL
          versions:
            - 6
            - 7
    dependencies: []

    ## local-role-base/role1/tasks/main.yml
    - ping:

    ## local-role-base/role1/tests/docker-compose.EL.6.yml
    version: "2"
    services:
      default:
        image: "centos:centos6"
        tty: True

    ## local-role-base/role1/tests/docker-compose.EL.7.yml
    version: "2"
    services:
      default:
        image: "centos:centos7"
        tty: True

    ## local-role-base/role1/tests/inventory
    default ansible_user=root

    ## local-role-base/role1/tests/test_playbook.yml
    - hosts: default
      gather_facts: no
      tasks:
        - name: host is reachable
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


@skip_if_no_docker
def test_hosts_without_domain_can_resolve_each_other(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthost1:
        image: "busybox:latest"
        tty: True

      guesthost2:
        image: "busybox:latest"
        tty: True

    ## inventory
    guesthost1 ansible_user=root
    guesthost2 ansible_user=root

    ## test_playbook.yml
    - hosts: guesthost1
      gather_facts: no
      tasks:
        - name: assert guesthost2 is reachable
          raw: ping -c 1 guesthost2
          changed_when: False
          tags: test

    - hosts: guesthost2
      gather_facts: no
      tasks:
        - name: assert guesthost1 is reachable
          raw: ping -c 1 guesthost1
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


@skip_if_no_docker
def test_hosts_with_domain_can_resolve_each_other(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      host1.domain1.tld:
        image: "busybox:latest"
        tty: True
        domainname: domain1.tld
        hostname: host1
        dns_search:
          - domain1.tld

      host2.domain1.tld:
        image: "busybox:latest"
        tty: True
        domainname: domain1.tld
        hostname: host2
        dns_search:
          - domain1.tld

      host1.domain2.tld:
        image: "busybox:latest"
        tty: True
        domainname: domain2.tld
        hostname: host1
        dns_search:
          - domain2.tld

    ## inventory
    host1.domain1.tld ansible_user=root
    host2.domain1.tld ansible_user=root
    host1.domain2.tld ansible_user=root

    ## test_playbook.yml
    - hosts: host1.domain1.tld
      gather_facts: no
      tasks:
        # host1.domain1.tld assertions
        - name: assert host1.domain1.tld can be resolved on host1.domain1.tld
          raw: ping -c 1 host1.domain1.tld
          changed_when: False
          tags: test

        - name: assert host1 can be resolved on host1.domain1.tld
          raw: ping -c 1 host1
          changed_when: False
          tags: test

        # host2.domain1.tld assertions
        - name: assert host2.domain1.tld can be resolved on host1.domain1.tld
          raw: ping -c 1 host2.domain1.tld
          changed_when: False
          tags: test

        - name: assert host2 can be resolved on host1.domain1.tld
          raw: ping -c 1 host2
          changed_when: False
          tags: test

        # host1.domain2.tld assertions
        - name: assert host1.domain2.tld can be resolved on host1.domain1.tld
          raw: ping -c 1 host1.domain2.tld
          changed_when: False
          tags: test

    - hosts: host2.domain1.tld
      gather_facts: no
      tasks:
        # host1.domain1.tld assertions
        - name: assert host1.domain1.tld can be resolved on host2.domain1.tld
          raw: ping -c 1 host1.domain1.tld
          changed_when: False
          tags: test

        - name: assert host1 can be resolved on host2.domain1.tld
          raw: ping -c 1 host1
          changed_when: False
          tags: test

        # host2.domain1.tld assertions
        - name: assert host2.domain1.tld can be resolved on host2.domain1.tld
          raw: ping -c 1 host2.domain1.tld
          changed_when: False
          tags: test

        - name: assert host2 can be resolved on host2.domain1.tld
          raw: ping -c 1 host2
          changed_when: False
          tags: test

        # host1.domain2.tld assertions
        - name: assert host1.domain2.tld can be resolved on host2.domain1.tld
          raw: ping -c 1 host1.domain2.tld
          changed_when: False
          tags: test

    - hosts: host1.domain2.tld
      gather_facts: no
      tasks:
        # host1.domain1.tld assertions
        - name: assert host1.domain1.tld can be resolved on host1.domain2.tld
          raw: ping -c 1 host1.domain1.tld
          changed_when: False
          tags: test

        # host2.domain1.tld assertions
        - name: assert host2.domain1.tld can be resolved on host1.domain2.tld
          raw: ping -c 1 host2.domain1.tld
          changed_when: False
          tags: test

        # host1.domain2.tld assertions
        - name: assert host1.domain2.tld can be resolved on host1.domain2.tld
          raw: ping -c 1 host1.domain2.tld
          changed_when: False
          tags: test

        - name: assert host1 can be resolved on host1.domain2.tld
          raw: ping -c 1 host1
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=14)


@skip_if_no_docker
def test_hosts_can_resolve_google_com_domain(testdir):
    smart_create(testdir.tmpdir, '''
    ## docker-compose.yml
    version: "2"
    services:
      guesthost:
        image: "busybox:latest"
        tty: True

    ## inventory
    guesthost ansible_user=root

    ## test_playbook.yml
    - hosts: guesthost
      gather_facts: no
      tasks:
        - name: assert google.com domain is reachable
          raw: ping -c 1 google.com
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


@skip_if_no_docker
def test_docker_build_error_results_in_fail_with_message(testdir):
    smart_create(testdir.tmpdir, '''
    ## image/Dockerfile
    FROM "https://unknownregistry/busybox:latest"

    ## docker-compose.yml
    version: "2"
    services:
      guesthost:
        build:
          context: "./image"
        tty: True

    ## inventory
    guesthost ansible_user=root

    ## test_playbook.yml
    - hosts: guesthost
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)

    assert "Failed: building service 'guesthost' failed with reason 'invalid reference format'" \
        in str(result.getfailures()[0].longrepr)



# -*- coding: utf-8 -*-

from __future__ import print_function

import io
import tarfile
import textwrap

import docker
import pytest
import requests


def is_docker_available():
    docker_client = docker.from_env()

    try:
        docker_client.version(api_version=False)
        return True
    except requests.exceptions.ConnectionError:
        return False


skip_if_no_docker = pytest.mark.skipif(
    not is_docker_available(),
    reason='docker is not available')


def smart_create(base_path, smart_content):
    smart_content = textwrap.dedent(smart_content)

    print(smart_content)

    for rel_path, content in smart_content_iter(smart_content, path_prefix='## '):
        path = base_path.join(*rel_path.split('/'))

        if path.basename.endswith('.tar.gz'):
            path.dirpath().ensure(dir=True)

            with tarfile.open(str(path), 'w:gz') as tar:
                for rel_tar_path, tar_content in smart_content_iter(content, path_prefix='#### '):
                    tar_bytes_content = tar_content.encode('utf-8')
                    info = tarfile.TarInfo(rel_tar_path)
                    info.size = len(tar_bytes_content)
                    tar.addfile(info, io.BytesIO(tar_bytes_content))
        else:
            path.write(content, ensure=True)


def smart_content_iter(smart_content, path_prefix):
    rel_path = None
    content = []

    for line in smart_content.splitlines():
        if line.startswith(path_prefix):
            if rel_path:
                yield rel_path, '\n'.join(content)
            rel_path = line[len(path_prefix):]
            content = []
        else:
            content.append(line)

    if rel_path:
        yield rel_path, '\n'.join(content)



# -*- coding: utf-8 -*-

import pytest

from goodplay_helpers import smart_create

pytestmark = pytest.mark.integration


def test_nodeid_no_environment(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: all
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].nodeid == 'test_playbook.yml::task1'


def test_nodeid_with_environment(testdir):
    smart_create(testdir.tmpdir, '''
    ## role1/meta/main.yml
    galaxy_info:
      author: John Doe
      platforms:
        - name: EL
          versions:
            - 6
            - 7
    dependencies: []

    ## role1/tasks/main.yml
    - ping:

    ## role1/tests/docker-compose.EL.6.yml
    version: "2"
    services:
      default:
        image: centos:centos6
        tty: True

    ## role1/tests/docker-compose.EL.7.yml
    version: "2"
    services:
      default:
        image: centos:centos7
        tty: True

    ## role1/tests/inventory
    default ansible_user=root

    ## role1/tests/test_playbook.yml
    - hosts: default
      tasks:
        - name: host is reachable
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 2
    assert items[0].nodeid == 'role1/tests/test_playbook.yml::EL.6::host is reachable'
    assert items[1].nodeid == 'role1/tests/test_playbook.yml::EL.7::host is reachable'



# -*- coding: utf-8 -*-

import logging

import pytest

from goodplay_helpers import smart_create

pytestmark = pytest.mark.integration


def test_goodplay_info_is_logged_to_stdout_and_logging(testdir, caplog, capfd):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    testdir.inline_run('-s')

    message = 'soft dependencies file not found at {0!s} ... nothing to install'.format(
        testdir.tmpdir.join('requirements.yml'))

    # capfd needs to be used here as logging module seems to use fd
    stdout, _ = capfd.readouterr()
    # assert message is exactly once in stdout (no double logging occurs)
    assert stdout.count(message) == 1

    assert ('goodplay.ansible_support.playbook', logging.INFO, message) in caplog.record_tuples


def test_goodplay_info_is_logged_to_stdout_when_logging_is_not_configured(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.runpytest_subprocess('-s')

    assert len([line for line in result.stdout.lines
                if 'soft dependencies file not found at ' in line]) == 1


def test_passed_on_non_changed_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_multiple_on_previously_changed_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          changed_when: True
          tags: test

        - name: task2
          ping:
          tags: test

        - name: task3
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2, failed=1)


def test_passed_on_previously_changed_non_test_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentionally changed task
          ping:
          changed_when: True

        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_on_previously_skipped_non_test_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentionally skipped task
          ping:
          when: False

        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_on_previously_failed_non_test_task_with_ignore_errors(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentionally failed task
          ping:
          failed_when: True
          ignore_errors: yes

        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_on_multiple_plays(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task2
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


def test_passed_with_gather_facts_enabled(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: yes
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_with_gather_facts_disabled(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_skipped_outcome_takes_priority_over_passed(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    host1 ansible_connection=local
    host2 ansible_connection=local

    ## test_playbook.yml
    - hosts: host1
      gather_facts: no
      tasks:
        - name: avoid all test tasks skipped
          ping:
          tags: test

    - hosts: host1:host2
      gather_facts: no
      tasks:
        - name: task2
          ping:
          when: inventory_hostname != 'host2'
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1, skipped=1)


def test_skipped_on_previously_failed_non_test_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentionally failed task
          ping:
          failed_when: True

        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(skipped=1, failed=1)


def test_skipped_multiple_on_previously_failed_non_test_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentionally failed task
          ping:
          failed_when: True

        - name: task1
          ping:
          tags: test

        - name: task2
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(skipped=2, failed=1)


def test_failed_on_changed_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          changed_when: True
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_failed_on_failed_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          failed_when: True
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_passed_on_wait_for_success(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: create .pid file
          file:
            path: "{{ playbook_dir }}/.pid"
            state: touch

        - name: task1
          wait_for:
            path: "{{ playbook_dir }}/.pid"
            timeout: 5
          tags: test

        - name: task2
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2)


def test_passed_on_include_playbook(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## playbook_to_include.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: create .pid file
          file:
            path: "{{ playbook_dir }}/.pid"
            state: touch

    ## test_playbook.yml
    - include: playbook_to_include.yml

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          wait_for:
            path: "{{ playbook_dir }}/.pid"
            timeout: 5
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_passed_on_import_playbook(testdir):
    pytest.importorskip('ansible.release', minversion='2.4')

    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## playbook_to_import.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: create .pid file
          file:
            path: "{{ playbook_dir }}/.pid"
            state: touch

    ## test_playbook.yml
    - import_playbook: playbook_to_import.yml

    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          wait_for:
            path: "{{ playbook_dir }}/.pid"
            timeout: 5
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_failed_on_wait_for_timeout(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          wait_for:
            path: /path/to/some/nonexisting/file
            timeout: 0
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_wait_for_task_timeout_does_not_stop_test_run(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          wait_for:
            port: 143
            timeout: 1
          tags: test

        - name: task2
          ping:
          tags: test

        - name: task3
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2, failed=1)


def test_wait_for_task_timeout_does_not_stop_test_run_with_multiple_hosts(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    localhost1 ansible_connection=local
    localhost2 ansible_connection=local

    ## test_playbook.yml
    - hosts: localhost1:localhost2
      gather_facts: no
      tasks:
        - name: task1
          wait_for:
            port: 143
            timeout: 1
          tags: test

        - name: task2
          ping:
          tags: test

        - name: task3
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=2, failed=1)


def test_failed_on_unreachable_host_on_gather_facts(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    unreachable.host.local

    ## test_playbook.yml
    - hosts: unreachable.host.local
      gather_facts: yes
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(skipped=1, failed=1)


def test_failed_on_single_failed_host(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    host1 ansible_connection=local
    host2 ansible_connection=local

    ## test_playbook.yml
    - hosts: host1:host2
      gather_facts: no
      tasks:
        - name: task1
          ping:
          failed_when: inventory_hostname == 'host1'
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_failed_on_all_tasks_skipped(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: unknownhost
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(skipped=1, failed=1)
    assert 'Failed: all test tasks have been skipped' in \
        str(result.getfailures()[0].longrepr)


def test_failed_outcome_takes_highest_priority(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    host1 ansible_connection=local
    host2 ansible_connection=local
    host3 ansible_connection=local

    ## test_playbook.yml
    - hosts: host1:host2:host3
      gather_facts: no
      tasks:
        - name: task1
          ping:
          when: inventory_hostname != 'host2'
          failed_when: inventory_hostname == 'host3'
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)


def test_failed_on_failing_non_test_task_after_passed_test_task(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test

        - name: intentionally failed task
          ping:
          failed_when: True
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1, failed=1)


def test_ansible_stdout_is_fully_consumed(testdir, capsys):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test

        - name: intentionally failed task
          ping:
          failed_when: True
    ''')

    testdir.inline_run('-s')
    stdout, _ = capsys.readouterr()
    ansible_play_recap_part = 'failed=1'
    assert ansible_play_recap_part in stdout


def test_ansible_retry_files_are_disabled(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          failed_when: True
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)

    assert not testdir.tmpdir.join('test_playbook.retry').check()


def test_goodplay_traceback_is_absent(testdir, capsys):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: task1
          ping:
          tags: test

        - name: intentionally failed task
          ping:
          failed_when: True

    ## test_something.py
    def please_fail():
        raise Exception()

    def test_something():
        please_fail()
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1, failed=2)

    stdout, _ = capsys.readouterr()
    stacktrace_entry_separator_count = \
        len([line for line in stdout.splitlines() if line.startswith('_ _ _ _ ')])

    # assert only single stacktrace entry separator for failing Python test
    assert stacktrace_entry_separator_count == 1



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import logging
import sarge

log = logging.getLogger(__name__)


def run(command, *args, **kwargs):
    command = sarge.shell_format(command, *args)
    log.info('run process: %s', command)

    return sarge.run(command, stdout=Capture(), stderr=Capture(), **kwargs)


class Capture(sarge.Capture):
    def __iter__(self):
        # override default Capture to read lines as long as streams are open,
        # thus iteration is not being stopped by large pauses between lines
        # being available
        while self.streams_open():
            line = self.readline()
            if not line:
                continue
            yield line.decode('utf-8')

        # ensure remaining buffered lines are consumed (original code)
        while True:
            line = self.readline()
            if not line:
                break
            yield line.decode('utf-8')



# -*- coding: utf-8 -*-

from goodplay.ansible_support import Inventory, Playbook
from goodplay.context import GoodplayContext


# inventory_path

def test_inventory_path_file_found_in_playbook_path_dir(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    inventory_path = tmpdir.join('inventory')
    inventory_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory_path == inventory_path


def test_inventory_path_dir_found_in_playbook_path_dir(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    inventory_path = tmpdir.join('inventory')
    inventory_path.ensure(dir=True)

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory_path == inventory_path


def test_inventory_path_is_none_when_not_in_playbook_path_dir(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory_path is None


def test_inventory_path_is_none_when_inventory_path_in_parent_dir(tmpdir):
    playbook_path = tmpdir.join('parent_dir', 'test_playbook.yml')
    playbook_path.ensure()

    inventory_path = tmpdir.join('inventory')
    inventory_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory_path is None


def test_inventory_path_is_none_when_inventory_path_in_sub_dir(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    inventory_path = tmpdir.join('sub_dir', 'inventory')
    inventory_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory_path is None


def test_inventory_path_is_cached(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    inventory_path = tmpdir.join('inventory')
    inventory_path.ensure()

    ctx = GoodplayContext(playbook_path)

    first_result = ctx.inventory_path
    assert first_result is not None
    assert id(ctx.inventory_path) == id(first_result)


# inventory

def test_inventory_is_none_when_inventory_path_is_none(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)
    ctx.inventory_path = None

    assert ctx.inventory is None


def test_inventory_class(tmpdir):
    inventory_path = tmpdir.join('inventory')
    inventory_path.write('host1', ensure=True)

    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.inventory.__class__ == Inventory


def test_inventory_is_cached(tmpdir):
    inventory_path = tmpdir.join('inventory')
    inventory_path.write('host1', ensure=True)

    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    first_result = ctx.inventory
    assert first_result is not None
    assert id(ctx.inventory) == id(first_result)


# playbook_dir_path

def test_playbook_dir_path_is_playbook_path_directory(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.playbook_dir_path == tmpdir


def test_playbook_dir_path_is_cached(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    first_result = ctx.playbook_dir_path
    assert first_result is not None
    assert id(ctx.playbook_dir_path) == id(first_result)


# playbook

def test_playbook_is_none_when_inventory_path_is_none(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)
    ctx.inventory_path = None

    assert ctx.playbook is None


def test_playbook_class(tmpdir):
    inventory_path = tmpdir.join('inventory')
    inventory_path.write('host1', ensure=True)

    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.write('''---
- hosts: host1
  tasks:
    - name: task 1
      ping:
      tags: test
''', ensure=True)

    ctx = GoodplayContext(playbook_path)

    assert ctx.playbook.__class__ == Playbook


def test_playbook_is_cached(tmpdir):
    inventory_path = tmpdir.join('inventory')
    inventory_path.write('host1', ensure=True)

    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.write('''---
- hosts: host1
  tasks:
    - name: task 1
      ping:
      tags: test
''', ensure=True)

    ctx = GoodplayContext(playbook_path)

    first_result = ctx.playbook
    assert first_result is not None
    assert id(ctx.playbook) == id(first_result)


# role_path

def test_role_path_is_none_when_plain_test_playbook(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path is None


def test_role_path_is_none_when_tests_parent_dir_only(tmpdir):
    playbook_path = tmpdir.join('tests', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path is None


def test_role_path_is_none_when_meta_dir_beside_tests_parent_dir_only(tmpdir):
    tmpdir.join('meta').ensure(dir=True)

    playbook_path = tmpdir.join('tests', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path is None


def test_role_path_is_none_when_meta_dir_with_main_yml_dir_beside_tests_parent_dir(tmpdir):
    tmpdir.join('meta', 'main.yml').ensure(dir=True)

    playbook_path = tmpdir.join('tests', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path is None


def test_role_path_is_none_when_meta_dir_with_main_yml_file_beside_playbook_path_dir(tmpdir):
    tmpdir.join('meta', 'main.yml').ensure()

    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path is None


def test_role_path_when_meta_dir_with_main_yml_file_beside_tests_parent_dir(tmpdir):
    tmpdir.join('meta', 'main.yml').ensure()

    playbook_path = tmpdir.join('tests', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path == tmpdir


def test_role_path_when_meta_dir_with_main_yml_file_beside_tests_ancestor_dir(tmpdir):
    tmpdir.join('meta', 'main.yml').ensure()

    playbook_path = tmpdir.join('tests', 'sub_dir1', 'sub_dir2', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.role_path == tmpdir


def test_role_path_is_cached(tmpdir):
    tmpdir.join('meta', 'main.yml').ensure()

    playbook_path = tmpdir.join('tests', 'test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    first_result = ctx.role_path
    assert first_result is not None
    assert id(ctx.role_path) == id(first_result)


# is_role_playbook

def test_is_role_playbook_is_false_when_role_path_is_none(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)
    ctx.role_path = None

    assert ctx.is_role_playbook is False


def test_is_role_playbook_is_true_when_role_path_is_not_none(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)
    ctx.role_path = tmpdir

    assert ctx.is_role_playbook is True


# compose_project_name

def test_compose_project_name_is_same_for_playbook_path_and_environment(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.compose_project_name('env1') == ctx.compose_project_name('env1')


def test_compose_project_name_incorporates_node_id(tmpdir, monkeypatch):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    compose_project_name1 = ctx.compose_project_name('env1')

    monkeypatch.setattr('uuid.getnode', lambda: 1234)

    compose_project_name2 = ctx.compose_project_name('env1')

    assert compose_project_name1 != compose_project_name2


def test_compose_project_name_incorporates_playbook_path(tmpdir):
    playbook_path1 = tmpdir.join('dir1', 'test_playbook.yml')
    playbook_path1.ensure()

    playbook_path2 = tmpdir.join('dir2', 'test_playbook.yml')
    playbook_path2.ensure()

    ctx1 = GoodplayContext(playbook_path1)
    ctx2 = GoodplayContext(playbook_path2)

    assert ctx1.compose_project_name('env1') != ctx2.compose_project_name('env1')


def test_compose_project_name_incorporates_environment(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)

    assert ctx.compose_project_name('env1') != ctx.compose_project_name('env2')


# release

def test_release_removes_temp_dir_paths(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')
    playbook_path.ensure()

    ctx = GoodplayContext(playbook_path)
    temp_dir_path = ctx._create_temp_dir_path()

    assert temp_dir_path.check(dir=True)
    assert len(ctx._temp_dir_paths) == 1

    ctx.release()

    assert not temp_dir_path.check()
    assert len(ctx._temp_dir_paths) == 0



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import logging
import sys

from cached_property import cached_property
import pytest

from goodplay import ansible_support, docker_support, junitxml
from goodplay.context import GoodplayContext

junitxml.patch_pytest_to_strip_file_extensions()


# https://urllib3.readthedocs.org/en/latest/security.html#insecureplatformwarning
logging.captureWarnings(True)


def enable_logging_goodplay_info_to_stdout():
    goodplay_stdout_handler = logging.StreamHandler(sys.stdout)
    goodplay_stdout_handler.setLevel(logging.INFO)
    goodplay_stdout_handler.setFormatter(logging.Formatter())

    goodplay_logger = logging.getLogger('goodplay')
    goodplay_logger.addHandler(goodplay_stdout_handler)
    goodplay_logger.setLevel(logging.DEBUG)


enable_logging_goodplay_info_to_stdout()


def pytest_addoption(parser):
    parser.getgroup('goodplay').addoption(
        '--use-local-roles', dest='use_local_roles', action='store_true',
        help='prefer to use local roles instead of auto-installed requirements')


class GoodplayFailed(Exception):
    pass


@pytest.mark.hookwrapper
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()

    if call.excinfo and call.excinfo.typename == 'GoodplayFailed':
        report.longrepr = 'Failed: {}'.format(call.excinfo.value)


# - GoodplayPlaybookFile (pytest.File)
#   - GoodplayEnvironment (pytest.Collector) -- manage Docker
#     - GoodplayPlaybook (pytest.Collector) -- manage Ansible runner
#       - GoodplayTest (pytest.Item)


def pytest_collect_file(parent, path):
    return GoodplayPlaybookFile.consider_and_create(path, parent)


class GoodplayContextSupport(object):
    @cached_property
    def ctx(self):
        return self.parent.ctx


# generic playbook preparations
class GoodplayPlaybookFile(pytest.File):
    def __init__(self, ctx, fspath, parent=None, config=None, session=None):
        super(GoodplayPlaybookFile, self).__init__(fspath, parent, config, session)

        self.ctx = ctx

    @classmethod
    def consider_and_create(cls, path, parent):
        if ansible_support.is_test_playbook_file(path):
            ctx = GoodplayContext(playbook_path=path, config=parent.config)

            if ctx.inventory_path:
                return GoodplayPlaybookFile(ctx, path, parent)

    def collect(self):
        try:
            environment_names = \
                docker_support.environment_names_for_playbook_path(self.ctx.playbook_path)

            if environment_names:
                for environment_name in environment_names:
                    yield GoodplayEnvironment(environment_name, self, self.config, self.session)
            else:
                yield GoodplayEnvironment(None, self, self.config, self.session)
        finally:
            if self.config.option.collectonly:
                self.ctx.release()

    def teardown(self):
        self.ctx.release()


# environment can be unspecific
class GoodplayEnvironment(GoodplayContextSupport, pytest.Collector):
    def __init__(self, environment_name, parent=None, config=None, session=None):
        if environment_name:
            # let the super class calculate the node id
            nodeid = None
        else:
            nodeid = parent.nodeid

        super(GoodplayEnvironment, self).__init__(
            environment_name, parent, config, session, nodeid=nodeid)
        self.environment_name = environment_name

        self.docker_runner = None

    def collect(self):
        yield GoodplayPlaybook(self.parent.name, self, self.config, self.session)

    def setup(self):
        self.docker_runner = docker_support.DockerRunner(self.ctx, self.environment_name)
        self.docker_runner.setup()

    def teardown(self):
        if self.docker_runner:
            self.docker_runner.teardown()


# environment specific playbook preparations
class GoodplayPlaybook(GoodplayContextSupport, pytest.Collector):
    def __init__(self, name, parent=None, config=None, session=None):
        super(GoodplayPlaybook, self).__init__(name, parent, config, session, nodeid=parent.nodeid)

        self.playbook_runner = None

    def collect(self):
        for task in self.ctx.playbook.test_tasks:
            yield GoodplayTest(task, self)

    def setup(self):
        self.playbook_runner = self.ctx.playbook.create_runner()
        self.playbook_runner.run_async()

    def teardown(self):
        if self.playbook_runner:
            self.playbook_runner.wait()

            if self.playbook_runner.failures:
                raise GoodplayFailed('\n'.join(self.playbook_runner.failures))


class GoodplayTest(GoodplayContextSupport, pytest.Item):
    def __init__(self, task, parent=None, config=None, session=None):
        super(GoodplayTest, self).__init__(task.name, parent, config, session)

        self.task = task

    def __repr__(self):
        return "<GoodplayTest '{0}'>".format(self.name)

    @cached_property
    def playbook_runner(self):
        return self.parent.playbook_runner

    def setup(self):
        self.playbook_runner.wait_for_test_task(self.task)

    def runtest(self):
        outcome = self.playbook_runner.wait_for_test_task_outcome(self.task)

        if outcome in ('skipped', None):
            pytest.skip()
        elif outcome == 'failed':
            raise GoodplayFailed()



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.utils.vars import load_extra_vars

from ansible.vars.manager import VariableManager


class EmptyOptions(object):
    extra_vars = []


class Inventory(object):
    def __init__(self, inventory_path):
        self.inventory_path = inventory_path
        self.inventory = None
        self.inventory = self.build_inventory()

    def build_inventory(self):
        self.clear_caches()

        loader = DataLoader()
        inventory = InventoryManager(
            loader=loader, sources=self.inventory_path.strpath)
        variable_manager = VariableManager(
            loader=loader, inventory=inventory)
        variable_manager.extra_vars = load_extra_vars(
            loader=loader, options=EmptyOptions())

        return inventory

    def clear_caches(self):
        if self.inventory is not None:
            self.inventory.clear_caches()

    def hosts(self):
        self.clear_caches()

        return [Host(host) for host in self.inventory.get_hosts()]


class Host(object):
    def __init__(self, host):
        self.host = host

    def vars(self):
        return self.host.get_vars()



# -*- coding: utf-8 -*-

from goodplay_helpers import smart_create


def test_ansible_error_message_is_forwarded(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          unknownmodule:
          tags: test
    ''')

    _, result = testdir.inline_genitems()
    result.assertoutcome(failed=1)

    assert result.getfailures()[0].__class__.__name__ == 'CollectReport'

    assert 'ERROR! no action detected in task' \
        in str(result.getfailures()[0].longrepr)


def test_nothing_collected_when_inventory_missing(testdir):
    smart_create(testdir.tmpdir, '''
    ## test_playbook.yml
    - hosts: all
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_nothing_collected_when_only_non_test_tags(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: other
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_fail_on_non_unique_task_names_both_tagged_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test

        - name: task1
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome(failed=1)

    message = "ValueError: Playbook '{0!s}' contains tests with non-unique name 'task1'".format(
        testdir.tmpdir.join('test_playbook.yml'))
    assert message in str(result.getfailures()[0].longrepr)
    assert len(items) == 0


def test_pass_on_non_unique_task_names_single_one_tagged_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:

        - name: task1
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task1'


def test_pass_on_tagged_test_with_additional_tags(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags:
            - other1
            - test
            - other2
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task1'


def test_single_play_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_single_play_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task1'


def test_single_play_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test

        - name: task2
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 2
    assert items[0].name == 'task1'
    assert items[1].name == 'task2'


def test_multiple_plays_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:

    - hosts: 127.0.0.1
      tasks:
        - name: task2
          ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_multiple_plays_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:

    - hosts: 127.0.0.1
      tasks:
        - name: task2
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task2'


def test_multiple_plays_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test

        - name: task2
          ping:

    - hosts: 127.0.0.1
      tasks:
        - name: task3
          ping:

        - name: task4
          ping:
          tags: test

        - name: task5
          ping:
          tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 3
    assert items[0].name == 'task1'
    assert items[1].name == 'task4'
    assert items[2].name == 'task5'


# ansible block support

def test_single_play_single_block_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_single_play_single_block_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task1'


def test_single_play_single_block_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
              tags: test

            - name: task2
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 2
    assert items[0].name == 'task1'
    assert items[1].name == 'task2'


def test_single_play_multiple_blocks_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

        - block:
            - name: task2
              ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_single_play_multiple_blocks_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

        - block:
            - name: task2
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task2'


def test_single_play_multiple_blocks_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
              tags: test

            - name: task2
              ping:

        - block:
            - name: task3
              ping:
              tags: test

            - name: task4
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 3
    assert items[0].name == 'task1'
    assert items[1].name == 'task3'
    assert items[2].name == 'task4'


def test_multiple_plays_single_block_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task2
              ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_multiple_plays_single_block_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task2
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task2'


def test_multiple_plays_single_block_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

            - name: task2
              ping:
              tags: test

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task3
              ping:

            - name: task4
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 2
    assert items[0].name == 'task2'
    assert items[1].name == 'task4'


def test_multiple_plays_multiple_blocks_no_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

        - block:
            - name: task2
              ping:

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task3
              ping:

        - block:
            - name: task4
              ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_multiple_plays_multiple_blocks_single_test(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:

        - block:
            - name: task2
              ping:

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task3
              ping:

        - block:
            - name: task4
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 1
    assert items[0].name == 'task4'


def test_multiple_plays_multiple_blocks_multiple_tests(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
              tags: test

            - name: task2
              ping:

        - block:
            - name: task3
              ping:
              tags: test

            - name: task4
              ping:
              tags: test

    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task5
              ping:
              tags: test

        - block:
            - name: task6
              ping:
              tags: test

            - name: task7
              ping:
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 5
    assert [item.name for item in items] == ['task1', 'task3', 'task4', 'task5', 'task6']


def test_ignore_test_tasks_in_rescue_block(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
          rescue:
            - name: task2
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0


def test_ignore_test_tasks_in_always_block(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - block:
            - name: task1
              ping:
          rescue:
            - name: task2
              ping:
          always:
            - name: task3
              ping:
              tags: test
    ''')

    items, result = testdir.inline_genitems()
    result.assertoutcome()
    assert len(items) == 0



# -*- coding: utf-8 -*-

import subprocess

from goodplay_helpers import smart_create


def test_goodplay_runs_fine_without_any_warnings(tmpdir):
    smart_create(tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    stdout = subprocess.check_output(
        ['goodplay'], cwd=str(tmpdir))

    assert b'warning' not in stdout



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import json
import re

try:
    from ansible.plugins.loader import module_loader
except ImportError:
    from ansible.plugins import module_loader
from ansible.plugins.callback import CallbackBase


# run test tasks in check mode when supported

from ansible.playbook.play_context import PlayContext


class ActionInfo(object):
    def __init__(self):
        self._supports_check_mode_re = re.compile(b'\\bsupports_check_mode\\s*=\\s*True\\b')
        self._cache = {}

    def should_run_in_check_mode(self, action):
        return action not in ('command', 'shell') \
            and self.supports_check_mode(action)

    def supports_check_mode(self, action):
        is_supported = self._cache.get(action, None)
        not_found_in_cache = is_supported is None

        if not_found_in_cache:
            module_path = module_loader.find_plugin(action)

            if module_path:
                with open(module_path, 'rb') as module_file:
                    module_file_content = module_file.read()
                    is_supported = bool(self._supports_check_mode_re.search(module_file_content))
                    self._cache[action] = is_supported

        return is_supported


ActionInfo = ActionInfo()


def monkeypatch_play_context():
    original_set_task_and_variable_override = PlayContext.set_task_and_variable_override

    def goodplay_set_task_and_variable_override(self, task, *args, **kwargs):
        new_info = original_set_task_and_variable_override(self, task, *args, **kwargs)

        if 'test' in task.tags:
            # enable check mode if supported
            if ActionInfo.should_run_in_check_mode(task.action):
                new_info.check_mode = True

            # special task action handling
            if task.action == 'wait_for':
                # failing wait_for should not stop test execution, but should
                # pop up as test task fail due to change
                task.register = '_goodplay_wait_for_result'
                task.changed_when = '{{ _goodplay_wait_for_result.failed | default(False) }}'
                task.failed_when = False

        return new_info

    PlayContext.set_task_and_variable_override = goodplay_set_task_and_variable_override


monkeypatch_play_context()


# goodplay callback

class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'notification'
    CALLBACK_NAME = 'goodplay'

    def __init__(self, *args, **kwargs):
        super(CallbackModule, self).__init__(*args, **kwargs)

        self.previous_task = None
        self.previously_ended_task = None
        self.task = None

        self.reset_per_host_outcomes()

    # Ansible playbook-specific callback methods

    def v2_playbook_on_task_start(self, task, is_conditional):
        # ensure task is available as attribute on instance
        self.task = task

        self.check_and_handle_playbook_on_task_end()

        if self.is_test_task(self.task):
            self.playbook_on_test_task_start(self.task)

    def check_and_handle_playbook_on_task_end(self):
        if self.previous_task and self.previously_ended_task != self.previous_task:
            self.playbook_on_task_end(self.previous_task)
            self.previously_ended_task = self.previous_task

        self.previous_task = self.task

    def is_test_task(self, task):
        # task can be None for meta tasks (e.g. setup)
        return task and getattr(task, 'name') and 'test' in task.tags

    def playbook_on_test_task_start(self, task):
        self.send_event('test-task-start', name=task.name)

    def send_event(self, event_name, **kwargs):
        event_line_prefix = 'GOODPLAY => '
        event_json = json.dumps(dict(event_name=event_name, data=kwargs))
        msg = event_line_prefix + event_json
        self._display.display(msg)

    def playbook_on_task_end(self, task):
        if self.is_test_task(task):
            self.playbook_on_test_task_end(task)

    def playbook_on_test_task_end(self, task):
        outcome = self.final_test_task_outcome()

        self.send_event('test-task-end', name=task.name, outcome=outcome)
        self.reset_per_host_outcomes()

    def v2_playbook_on_play_start(self, play):
        self.check_and_handle_playbook_on_task_end()

    def v2_playbook_on_stats(self, stats):
        self.check_and_handle_playbook_on_task_end()

    # Ansible runner-specific callback methods

    def runner_on_failed(self, host, res, ignore_errors=False):
        if ignore_errors:
            return

        self.handle_failed_result(host, res)

    def runner_on_ok(self, host, res):
        changed = res.get('changed')

        if changed:
            self.handle_changed_result(host, res)
        else:
            self.handle_non_changed_result(host, res)

    def runner_on_skipped(self, host, item=None):
        self.handle_skipped_result(host)

    def runner_on_unreachable(self, host, res):
        self.handle_failed_result(host, res)

    # handle results and outcome

    def reset_per_host_outcomes(self):
        self.per_host_outcomes = {}

    def add_per_host_outcome(self, host, outcome, res=None):
        self.per_host_outcomes[host] = dict(outcome=outcome, res=res)

    def final_test_task_outcome(self):
        if self.per_host_outcomes:
            outcomes = set(x['outcome'] for x in self.per_host_outcomes.values())
            outcome_priority = ('failed', 'skipped', 'passed')
            for outcome in outcome_priority:
                if outcome in outcomes:
                    return outcome

        return 'skipped'

    def handle_changed_result(self, host, res):
        self.check_and_handle_test_task(host, 'failed', res)

    def handle_non_changed_result(self, host, res):
        self.check_and_handle_test_task(host, 'passed', res)

    def handle_skipped_result(self, host):
        self.check_and_handle_test_task(host, 'skipped')

    def handle_failed_result(self, host, res):
        if not self.check_and_handle_test_task(host, 'failed', res):
            self.send_event('error', message=str(res))

    def check_and_handle_test_task(self, host, outcome, res=None):
        if self.is_test_task(self.task):
            self.add_per_host_outcome(host, outcome, res)
            return True
        return False



# -*- coding: utf-8 -*-

from goodplay_helpers import smart_create


def test_file_task_tagged_with_test_runs_in_check_mode(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - file:
            path: "{{ playbook_dir }}/README"
            state: touch

        - name: intentionally failing test
          file:
            path: "{{ playbook_dir }}/README"
            state: absent
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)

    assert testdir.tmpdir.join('README').check(file=True)


def test_lineinfile_task_tagged_with_test_runs_in_check_mode(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - lineinfile:
            dest: "{{ playbook_dir }}/HELLO"
            create: yes
            line: "WORLD"
            state: present

        - name: intentionally failing test
          file:
            path: "{{ playbook_dir }}/HELLO"
            line: "WORLD"
            state: absent
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(failed=1)

    assert testdir.tmpdir.join('HELLO').check(file=True)
    assert testdir.tmpdir.join('HELLO').read() == 'WORLD\n'


def test_command_task_tagged_with_test_runs_in_normal_mode(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentional test with side effect
          command: touch "{{ playbook_dir }}/HELLO"
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)

    assert testdir.tmpdir.join('HELLO').check(file=True)


def test_shell_task_tagged_with_test_runs_in_normal_mode(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: intentional test with side effect
          shell: touch "{{ playbook_dir }}/HELLO"
          changed_when: False
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)

    assert testdir.tmpdir.join('HELLO').check(file=True)


def test_custom_module_runs_in_normal_mode_when_check_mode_not_supported(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## library/checkrunmode.py
    from ansible.module_utils.basic import AnsibleModule

    import os

    def main():
        module = AnsibleModule(
            argument_spec = dict(),
            supports_check_mode=False
        )

        module.exit_json(changed=False, run_in_check_mode=module.check_mode)

    if __name__ == '__main__':
        main()

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert custom module not run in check mode
          checkrunmode:
          register: result
          failed_when: "{{ result.run_in_check_mode }}"
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)


def test_custom_module_runs_in_check_mode_when_supported(testdir):
    smart_create(testdir.tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## library/checkrunmode.py
    from ansible.module_utils.basic import AnsibleModule

    import os

    def main():
        module = AnsibleModule(
            argument_spec = dict(),
            supports_check_mode=True
        )

        module.exit_json(changed=False, run_in_check_mode=module.check_mode)

    if __name__ == '__main__':
        main()

    ## test_playbook.yml
    - hosts: 127.0.0.1
      gather_facts: no
      tasks:
        - name: assert custom module run in check mode
          checkrunmode:
          register: result
          failed_when: "{{ not result.run_in_check_mode }}"
          tags: test
    ''')

    result = testdir.inline_run('-s')
    result.assertoutcome(passed=1)



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import pytest

from goodplay.ansible_support import is_test_playbook_file


class CollectOnlyTestPlaybooks(object):
    def pytest_ignore_collect(self, path, config):
        # do not ignore directories
        if path.check(dir=True):
            return False

        # ignore everything else that is not a test playbook
        return not is_test_playbook_file(path)


def main():
    additional_plugins = [CollectOnlyTestPlaybooks()]

    raise SystemExit(pytest.main(plugins=additional_plugins))



# -*- coding: utf-8 -*-

from goodplay.docker_support import (
    is_docker_compose_file, config_paths_for_playbook_path, environment_name_for_config_path)

from goodplay_helpers import smart_create


def test_is_docker_compose_file_returns_false_for_non_docker_compose_prefix(tmpdir):
    path = tmpdir.join('docker.yml')
    path.ensure()

    assert not is_docker_compose_file(path)


def test_is_docker_compose_file_returns_false_for_non_yml_suffix(tmpdir):
    path = tmpdir.join('docker-compose.json')
    path.ensure()

    assert not is_docker_compose_file(path)


def test_is_docker_compose_file_returns_false_for_directory(tmpdir):
    path = tmpdir.join('docker-compose.yml')
    path.ensure(dir=True)

    assert not is_docker_compose_file(path)


def test_is_docker_compose_file_returns_false_for_multidot_name(tmpdir):
    path = tmpdir.join('docker-compose..yml')
    path.ensure()

    assert not is_docker_compose_file(path)


def test_is_docker_compose_file_success(tmpdir):
    path = tmpdir.join('docker-compose.yml')
    path.ensure()

    assert is_docker_compose_file(path)


def test_environment_name_for_config_path_empty():
    assert environment_name_for_config_path([]) is None


def test_environment_name_for_config_path_docker_compose_yml_only():
    assert environment_name_for_config_path(['docker-compose.yml']) == ''


def test_environment_name_for_config_path_docker_compose_override_yml_only():
    assert environment_name_for_config_path(['docker-compose.override.yml']) == ''


def test_environment_name_for_config_path_docker_compose_yml_and_override():
    docker_compose_file_group = ['docker-compose.yml', 'docker-compose.override.yml']
    assert environment_name_for_config_path(docker_compose_file_group) == ''


def test_environment_name_for_config_path_single_level():
    docker_compose_file_group = ['docker-compose.yml', 'docker-compose.item1.yml']
    assert environment_name_for_config_path(docker_compose_file_group) == 'item1'


def test_environment_name_for_config_path_muliple_levels():
    docker_compose_file_group = [
        'docker-compose.yml',
        'docker-compose.item1.override.yml',
        'docker-compose.item1.item11.item111.override.yml'
    ]
    expected = 'item1.item11.item111'
    assert environment_name_for_config_path(docker_compose_file_group) == expected


def test_config_paths_for_playbook_path_empty(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    assert config_paths_for_playbook_path(playbook_path) == []


def test_config_paths_for_playbook_path_docker_compose_yml_only(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ''')

    expected = [['docker-compose.yml']]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_docker_compose_override_yml_only(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.override.yml
    ''')

    expected = [['docker-compose.override.yml']]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_docker_compose_yml_and_override(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.override.yml
    ''')

    expected = [
        ['docker-compose.yml', 'docker-compose.override.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_single_level(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.item1.yml
    ## docker-compose.item2.yml
    ''')

    expected = [
        ['docker-compose.item1.yml'],
        ['docker-compose.item2.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_single_level_only_override(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.item1.override.yml
    ''')

    expected = [
        ['docker-compose.yml', 'docker-compose.item1.override.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_single_level_and_override(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.item1.yml
    ## docker-compose.item1.override.yml
    ''')

    expected = [
        ['docker-compose.item1.yml', 'docker-compose.item1.override.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_multiple_levels(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.item1.item11.yml
    ## docker-compose.item2.item21.yml
    ''')

    expected = [
        ['docker-compose.item1.item11.yml'],
        ['docker-compose.item2.item21.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected


def test_config_paths_for_playbook_path_multiple_levels_and_override(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    smart_create(tmpdir, '''
    ## docker-compose.yml
    ## docker-compose.item1.override.yml
    ## docker-compose.item1.item11.item111.override.yml
    ## docker-compose.item2.item21.yml
    ''')

    expected = [
        ['docker-compose.yml', 'docker-compose.item1.override.yml',
         'docker-compose.item1.item11.item111.override.yml'],
        ['docker-compose.item2.item21.yml']
    ]
    assert config_paths_for_playbook_path(playbook_path) == expected



# -*- coding: utf-8 -*-

import subprocess

from goodplay_helpers import smart_create


def test_goodplay_cli_args_are_forwarded_to_pytest(tmpdir):
    smart_create(tmpdir, '''
    ## inventory
    127.0.0.1 ansible_connection=local

    ## test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    stdout = subprocess.check_output(
        ['goodplay', '--collect-only'], cwd=str(tmpdir))
    expected = b"<GoodplayTest 'task1'>"

    assert expected in stdout


def test_goodplay_cli_does_not_collect_python_tests(tmpdir):
    smart_create(tmpdir, '''
    ## test_somepython.py
    def test_something_python_related():
        pass
    ''')

    # pytest's exit code when no tests have been collected
    EXIT_NOTESTSCOLLECTED = 5

    goodplay_returncode = subprocess.call(
        ['goodplay'], cwd=str(tmpdir))

    assert goodplay_returncode == EXIT_NOTESTSCOLLECTED


def test_goodplay_cli_recursively_collects_from_subdirectories(tmpdir):
    smart_create(tmpdir, '''
    ## some/subdir/inventory
    127.0.0.1 ansible_connection=local

    ## some/subdir/test_playbook.yml
    - hosts: 127.0.0.1
      tasks:
        - name: task1
          ping:
          tags: test
    ''')

    stdout = subprocess.check_output(
        ['goodplay', '--collect-only'], cwd=str(tmpdir))
    expected = b"<GoodplayTest 'task1'>"

    assert expected in stdout


def test_goodplay_cli_forwards_pytest_returncode(tmpdir):
    # pytest's exit code when no tests have been collected
    EXIT_NOTESTSCOLLECTED = 5

    goodplay_returncode = subprocess.call(
        ['goodplay'], cwd=str(tmpdir))

    assert goodplay_returncode == EXIT_NOTESTSCOLLECTED



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import re

import _pytest.junitxml


def patch_pytest_to_strip_file_extensions():
    _pytest.junitxml._py_ext_re = re.compile(r'\.(?:py|yml)$')



# -*- coding: utf-8 -*-

from goodplay.plugin import (
    GoodplayPlaybookFile, GoodplayEnvironment, GoodplayPlaybook, GoodplayTest)


class AttrDict(dict):
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


def test_nodeid_goodplay_playbook_file(tmpdir):
    playbook_path = tmpdir.join('test_playbook.yml')

    config = AttrDict(rootdir=tmpdir)
    session = AttrDict(non='empty', config=config)

    node = GoodplayPlaybookFile(ctx=None, fspath=playbook_path, config=config, session=session)
    assert node.nodeid == 'test_playbook.yml'


def test_nodeid_goodplay_environment_with_no_environment_name_is_same_as_parent(tmpdir):
    parent = AttrDict(
        config=AttrDict(rootdir=tmpdir),
        session=AttrDict(non='empty'),
        fspath=tmpdir.join('test_playbook.yml'),
        nodeid='test_playbook.yml'
    )

    node = GoodplayEnvironment(environment_name=None, parent=parent)
    assert node.nodeid == 'test_playbook.yml'


def test_nodeid_goodplay_environment_with_environment_name_adds_environment_name(tmpdir):
    parent = AttrDict(
        config=AttrDict(rootdir=tmpdir),
        session=AttrDict(non='empty'),
        fspath=tmpdir.join('test_playbook.yml'),
        nodeid='test_playbook.yml'
    )

    node = GoodplayEnvironment(environment_name='ubuntu.trusty', parent=parent)
    assert node.nodeid == 'test_playbook.yml::ubuntu.trusty'


def test_nodeid_goodplay_environment_name_is_same_as_parent(tmpdir):
    parent = AttrDict(
        config=AttrDict(rootdir=tmpdir),
        session=AttrDict(non='empty'),
        fspath=tmpdir.join('test_playbook.yml'),
        nodeid='test_playbook.yml::ubuntu.trusty'
    )

    node = GoodplayPlaybook(name='test_playbook.yml', parent=parent)
    assert node.nodeid == 'test_playbook.yml::ubuntu.trusty'


def test_nodeid_goodplay_test_add_test_name(tmpdir):
    parent = AttrDict(
        config=AttrDict(rootdir=tmpdir),
        session=AttrDict(non='empty'),
        fspath=tmpdir.join('test_playbook.yml'),
        nodeid='test_playbook.yml::ubuntu.trusty'
    )

    task = AttrDict(name='host is reachable')

    node = GoodplayTest(task=task, parent=parent)
    assert node.nodeid == 'test_playbook.yml::ubuntu.trusty::host is reachable'



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import hashlib
import uuid

from cached_property import cached_property
import py.path

from goodplay import ansible_support


class GoodplayContext(object):
    def __init__(self, playbook_path, config=None):
        self.playbook_path = playbook_path
        self.config = config

        self._temp_dir_paths = []

    def _create_temp_dir_path(self):
        temp_path = py.path.local.mkdtemp()
        self._temp_dir_paths.append(temp_path)
        return temp_path

    @cached_property
    def inventory_path(self):
        inventory_path = self.playbook_path.dirpath('inventory')

        if inventory_path.check():
            return inventory_path

    @cached_property
    def inventory(self):
        if self.inventory_path:
            return ansible_support.Inventory(self.inventory_path)

    @cached_property
    def extended_inventory_path(self):
        extended_inventory_path = self._create_temp_dir_path()

        extended_inventory_path.join('inventory').mksymlinkto(self.inventory_path)

        if self.inventory_path.check(dir=True):
            group_vars_path = self.inventory_path.join('group_vars')

            if group_vars_path.check(dir=True):
                extended_inventory_path.join('group_vars').mksymlinkto(group_vars_path)

            host_vars_path = self.inventory_path.join('host_vars')

            if host_vars_path.check(dir=True):
                extended_inventory_path.join('host_vars').mksymlinkto(host_vars_path)

        # when inventory_path is a file, group_vars and host_vars directories
        # are handled implicitly as they are beside the executed playbook

        return extended_inventory_path

    @cached_property
    def playbook_dir_path(self):
        return self.playbook_path.dirpath()

    @cached_property
    def playbook(self):
        if self.inventory_path:
            return ansible_support.Playbook(self)

    @cached_property
    def is_role_playbook(self):
        return bool(self.role_path)

    @cached_property
    def role_path(self):
        for ancestor_path in self.playbook_path.parts(reverse=True)[1:]:
            if ancestor_path.basename == 'tests':
                role_path = ancestor_path.dirpath()
                is_role_path = role_path.join('meta', 'main.yml').check(file=True)

                if is_role_path:
                    return role_path
                break

    @cached_property
    def role_under_test_roles_path(self):
        role_under_test_roles_path = self._create_temp_dir_path()

        role_under_test_roles_path.join(self.role_path.basename).mksymlinkto(self.role_path)

        return role_under_test_roles_path

    @cached_property
    def installed_roles_path(self):
        return self._create_temp_dir_path()

    @cached_property
    def use_local_roles(self):
        return self.config.getoption('use_local_roles')

    def compose_project_name(self, environment_name):
        node_id = '{:x}'.format(uuid.getnode())
        project_name_parts = ':'.join((node_id, self.playbook_path.strpath, environment_name))
        project_hash = hashlib.sha1(project_name_parts.encode('utf-8')).hexdigest()[:8]

        return 'goodplay{}'.format(project_hash)

    def release(self):
        for temp_path in reversed(self._temp_dir_paths):
            temp_path.remove(ignore_errors=True)
        self._temp_dir_paths = []



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import collections
import logging
import os
import re

import ansible.constants
from cached_property import cached_property
import yaml

from .runner import PlaybookRunner
from ..utils.subprocess import run

log = logging.getLogger(__name__)


class Playbook(object):
    tagged_tasks_re = re.compile(r'''
        (?:^\ {6})        # 6 spaces at the beginning of line
        (?P<name>.*?)     # group: task name
        (?:\t)            # 1 tab delimiter
        TAGS:\ \[         # task tags prefix
        (?P<tags>.+?)     # group: task tags
        \]$               # task tags suffix at the end of line
        ''', re.VERBOSE)

    def __init__(self, ctx):
        self.ctx = ctx

        self.install_all_dependencies()

    def install_all_dependencies(self):
        self.install_role_dependencies()
        self.install_soft_dependencies()

    def install_role_dependencies(self):
        if not self.ctx.is_role_playbook:
            return

        role_meta_path = self.ctx.role_path.join('meta', 'main.yml')
        role_meta_content = yaml.safe_load(role_meta_path.read())
        role_dependencies = role_meta_content.get('dependencies', [])

        if role_dependencies:
            log.info('role dependencies found in %s ... installing', role_meta_path)
            requirements_path = self.ctx.installed_roles_path.join('requirements.yml')
            requirements_path.write(yaml.dump(role_dependencies))

            self.install_roles_from_requirements_file(requirements_path)
        else:
            log.info('role dependencies not found in %s ... nothing to install',
                     role_meta_path)

    def install_soft_dependencies(self):
        requirements_path = self.ctx.playbook_path.dirpath('requirements.yml')

        if requirements_path.check(file=True):
            log.info('soft dependencies found in %s ... installing', requirements_path)
            self.install_roles_from_requirements_file(requirements_path)
        else:
            log.info('soft dependencies file not found at %s ... nothing to install',
                     requirements_path)

    def install_roles_from_requirements_file(self, requirements_file):
        ansible_galaxy_cmd = \
            'ansible-galaxy install -vvvv --force --role-file {0} --roles-path {1}'

        if self.ctx.use_local_roles:
            ansible_galaxy_cmd += ' --ignore-errors'

        process = run(ansible_galaxy_cmd, requirements_file, self.ctx.installed_roles_path)

        for line in process.stdout:
            log.info(line[:-1])

        if process.returncode != 0:
            raise Exception(process.stderr.readlines())  # pragma: no cover

    def env(self):
        roles_path = []

        if self.ctx.role_path:
            roles_path.append(self.ctx.role_under_test_roles_path.strpath)

        if self.ctx.use_local_roles:
            roles_path.extend(ansible.constants.DEFAULT_ROLES_PATH)

        roles_path.append(self.ctx.installed_roles_path.strpath)

        return dict(
            ANSIBLE_ROLES_PATH=os.pathsep.join(roles_path),
            ANSIBLE_RETRY_FILES_ENABLED='false',
            )

    def create_runner(self):
        return PlaybookRunner(self.ctx)

    @cached_property
    def test_tasks(self):
        test_tasks = [task for task in self.tasks() if 'test' in task.tags]
        self.ensure_unique_task_names(test_tasks)

        return test_tasks

    def ensure_unique_task_names(self, tasks):
        non_unique_task_names = [
            task_name for task_name, count in
            collections.Counter(task.name for task in tasks).items()
            if count > 1]

        if non_unique_task_names:
            raise ValueError(
                "Playbook '{0}' contains tests with non-unique name '{1}'"
                .format(self.ctx.playbook_path.strpath, non_unique_task_names[0]))

    def tasks(self):
        process = run(
            'ansible-playbook --list-tasks --list-tags -i {0} {1}',
            self.ctx.inventory_path, self.ctx.playbook_path,
            env=self.env())

        if process.returncode != 0:
            raise Exception(process.stderr.read())

        for line in process.stdout:
            match = self.tagged_tasks_re.match(line)

            if match:
                name = match.group('name')
                tags = [tag.strip() for tag in match.group('tags').split(',')]

                yield Task(name, tags)


class Task(object):
    def __init__(self, name, tags):
        self.name = name
        self.tags = tags



# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages


install_requires = [
    'ansible>=2.5',
    'cached-property>=1.3.1',
    'docker-compose>=1.22.0',
    'py>=1.4.34',
    'pytest>=3.5.0',
    'sarge>=0.1.5',
]

readme_path = os.path.join(os.path.dirname(__file__), 'README.rst')
with open(readme_path, 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name='goodplay',
    version=__import__('goodplay').__version__,
    description='goodplay enables you to test your deployments and '
                'distributed software infrastructure by reusing your '
                'existing knowledge of Ansible.',
    long_description=long_description,

    author='Benjamin Schwarze',
    author_email='benjamin.schwarze@mailboxd.de',
    maintainer='Benjamin Schwarze',
    maintainer_email='benjamin.schwarze@mailboxd.de',

    url='https://goodplay.io/',

    license='Apache License 2.0',
    keywords=[
        'goodplay', 'good', 'play', 'test', 'deployment', 'ansible',
        'playbook', 'role', 'integration', 'system', 'tdd', 'configuration',
        'management'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
    ],

    entry_points={
        'console_scripts': [
            'goodplay = goodplay.cli:main',
        ],
        'pytest11': [
            'goodplay = goodplay.plugin',
        ]
    },
    packages=find_packages(),
    package_data={
        '': [
            'ansible_support/callback_plugin/goodplay.py',
            'ansible_support/connection_plugin/goodplaydocker.py',
        ],
    },
    install_requires=install_requires,
    zip_safe=False,
)



# -*- coding: utf-8 -*-

import pytest

from goodplay import ansible_support


# fixtures

@pytest.fixture
def valid_playbook_extension():
    return 'yml'


@pytest.fixture
def invalid_playbook_extension():
    return 'txt'


@pytest.fixture
def valid_test_playbook():
    return '''---
- hosts: default
  tasks:
    - name: host is reachable
      ping:
      tags: test
'''


@pytest.fixture(
    params=[
        '- hosts: default',
        '- include: abc',
        '- import_playbook: otherplaybook.yml',
    ],
    ids=[
        'hosts',
        'include',
        'import_playbook',
    ]
)
def valid_test_playbook_content(request):
    return request.param


@pytest.fixture(
    params=[
        '',
        'string',
        'key: value',
        '[]',
        '- hello world',
        '- key: value',
    ],
    ids=[
        'empty',
        'string',
        'dict',
        'empty-list',
        'non-dict-list',
        'non-play-dict-list',
    ]
)
def invalid_test_playbook_content(request):
    return request.param


# tests

# prefix tests

def test_is_test_playbook_file_with_valid_prefix(
        tmpdir, valid_test_playbook):
    path = tmpdir.join('test_playbook.yml')
    path.write(valid_test_playbook)

    assert ansible_support.is_test_playbook_file(path)


def test_is_test_playbook_file_with_invalid_prefix(
        tmpdir, valid_test_playbook):
    path = tmpdir.join('playbook.yml')
    path.write(valid_test_playbook)

    assert not ansible_support.is_test_playbook_file(path)


# extension tests

def test_is_test_playbook_file_with_valid_extension(
        tmpdir, valid_playbook_extension, valid_test_playbook):
    path = tmpdir.join('test_playbook.' + valid_playbook_extension)
    path.write(valid_test_playbook)

    assert ansible_support.is_test_playbook_file(path)


def test_is_test_playbook_file_with_invalid_extension(
        tmpdir, invalid_playbook_extension, valid_test_playbook):
    path = tmpdir.join('test_playbook.' + invalid_playbook_extension)
    path.write(valid_test_playbook)

    assert not ansible_support.is_test_playbook_file(path)


# content tests

def test_is_test_playbook_file_with_valid_content(
        tmpdir, valid_test_playbook_content):
    path = tmpdir.join('test_playbook.yml')
    path.write(valid_test_playbook_content)

    assert ansible_support.is_test_playbook_file(path)


def test_is_test_playbook_file_with_invalid_content(
        tmpdir, invalid_test_playbook_content):
    path = tmpdir.join('test_playbook.yml')
    path.write(invalid_test_playbook_content)

    assert not ansible_support.is_test_playbook_file(path)


def test_host_vars_are_not_mixed_when_using_multiple_inventories(tmpdir):
    inventory1_path = tmpdir.join('inventory1')
    inventory1_path.write('default key=value1')

    inventory2_path = tmpdir.join('inventory2')
    inventory2_path.write('default key=value2')

    inventory1 = ansible_support.Inventory(inventory1_path)
    inventory2 = ansible_support.Inventory(inventory2_path)

    inventory1_hosts = inventory1.hosts()
    inventory2_hosts = inventory2.hosts()

    assert len(inventory1_hosts) == 1
    assert inventory1_hosts[0].vars()['key'] == 'value1'

    assert len(inventory2_hosts) == 1
    assert inventory2_hosts[0].vars()['key'] == 'value2'



# -*- coding: utf-8 -*-

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import logging
import os

log = logging.getLogger(__name__)


def pytest_collectstart(collector):
    is_root_collector = collector.parent is None
    if is_root_collector:
        testdir_path = collector.fspath
        if testdir_path.check(dir=False):
            testdir_path = testdir_path.join(os.pardir)

        log_testdir_content(testdir_path)


def log_testdir_content(testdir_path):
    for path in testdir_path.visit(fil=lambda x: x.check(dir=False), bf=True, sort=True):
        with path.open(mode='rb') as f:
            is_binary_file = is_binary_string(f.read(1024))

            if is_binary_file:
                log.info('## %s (binary file, content skipped)', testdir_path.bestrelpath(path))
            else:
                log.info('## %s\n%s', testdir_path.bestrelpath(path), path.read())


# based on file(1) behavior, see http://stackoverflow.com/a/7392391/5611889
TEXTCHARS = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})


def is_binary_string(_bytes):
    return bool(_bytes.translate(None, TEXTCHARS))
